# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional

from dkd import ContentType, BaseContent


class TextContent(BaseContent):
    """
        Text Message Content
        ~~~~~~~~~~~~~~~~~~~~

        data format: {
            type : 0x01,
            sn   : 123,

            text : "..."
        }
    """

    def __init__(self, content: Optional[dict]=None, text: Optional[str]=None):
        if content is None:
            super().__init__(content_type=ContentType.TEXT)
        else:
            super().__init__(content=content)
        if text is not None:
            self.text = text

    #
    #   text
    #
    @property
    def text(self) -> str:
        return self.get('text')

    @text.setter
    def text(self, value: str):
        if value is None:
            self.pop('text', None)
        else:
            self['text'] = value



# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Document Command Protocol
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    1. contains 'ID' only, means query document for ID
    2. contains 'document' and 'signature' (must match), means reply
"""

from typing import Optional

from mkm import ID, Meta, Document

from .command import Command
from .meta import MetaCommand


class DocumentCommand(MetaCommand):
    """
        Document Command
        ~~~~~~~~~~~~~~~~

        data format: {
            type : 0x88,
            sn   : 123,

            command   : "document", // command name
            ID        : "{ID}",     // entity ID
            meta      : {...},      // only for handshaking with new friend
            document  : {...},      // when document is empty, means query for ID
            signature : "..."       // old document's signature for querying
        }

    """

    def __init__(self, cmd: Optional[dict]=None,
                 identifier: Optional[ID]=None, meta: Optional[Meta]=None, document: Optional[Document]=None,
                 signature: Optional[str]=None):
        if identifier is None and document is not None:
            identifier = document.identifier
        super().__init__(cmd, Command.DOCUMENT, identifier=identifier, meta=meta)
        self.__doc = document
        if document is not None:
            self['document'] = document.dictionary
        if signature is not None:
            self['signature'] = signature

    #
    #   document
    #
    @property
    def document(self) -> Optional[Document]:
        if self.__doc is None:
            data = self.get('profile')
            if isinstance(data, str):
                # compatible with v1.0
                #    "ID"        : "{ID}",
                #    "profile"   : "{JsON}",
                #    "signature" : "{BASE64}"
                dictionary = {
                    'ID': str(self.identifier),
                    'data': data,
                    'signature': self.get("signature")
                }
                data = dictionary
            else:
                if data is None:
                    data = self.get('document')
                # (v1.1)
                #    "ID"      : "{ID}",
                #    "profile" : {
                #        "ID"        : "{ID}",
                #        "data"      : "{JsON}",
                #        "signature" : "{BASE64}"
                #    }
                assert data is None or isinstance(data, dict), 'document data error: %s' % data
            self.__doc = Document.parse(document=data)
        return self.__doc

    @property
    def signature(self) -> Optional[str]:
        """
        signature for checking new document

        :return: part of signature in current document (base64)
        """
        return self.get('signature')

    @classmethod
    def query(cls, identifier: ID, signature: Optional[str]=None):
        return cls(identifier=identifier, signature=signature)

    @classmethod
    def response(cls, document: Document, meta: Optional[Meta]=None, identifier: Optional[ID]=None):
        return cls(identifier=identifier, meta=meta, document=document)



# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Meta Command Protocol
    ~~~~~~~~~~~~~~~~~~~~~

    1. contains 'ID' only, means query meta for ID
    2. contains 'meta' (must match), means reply
"""

from typing import Optional

from mkm import ID, Meta

from .command import Command


class MetaCommand(Command):
    """
        Meta Command
        ~~~~~~~~~~~~

        data format: {
            type : 0x88,
            sn   : 123,

            command : "meta", // command name
            ID      : "{ID}", // contact's ID
            meta    : {...}   // When meta is empty, means query meta for ID
        }
    """

    def __init__(self, cmd: Optional[dict]=None, command: Optional[str]=None,
                 identifier: Optional[ID]=None, meta: Optional[Meta]=None):
        if cmd is None:
            if command is None:
                command = Command.META
            super().__init__(command=command)
        else:
            super().__init__(cmd=cmd)
        self.__meta = meta
        if meta is not None:
            self['meta'] = meta.dictionary
        if identifier is not None:
            self['ID'] = str(identifier)

    #
    #   ID
    #
    @property
    def identifier(self) -> ID:
        return ID.parse(identifier=self.get('ID'))

    #
    #   Meta
    #
    @property
    def meta(self) -> Optional[Meta]:
        if self.__meta is None:
            self.__meta = Meta.parse(meta=self.get('meta'))
        return self.__meta

    @classmethod
    def query(cls, identifier: ID):
        return cls(identifier=identifier)

    @classmethod
    def response(cls, identifier: ID, meta: Meta):
        return cls(identifier=identifier, meta=meta)



# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import weakref
from abc import abstractmethod
from typing import Optional

from mkm import ID, Meta, Document


class EntityDataSource:
    """This interface is for getting information for entity(user/group)

        Entity Data Source
        ~~~~~~~~~~~~~~~~~~

        1. meta for user, which is generated by the user's private key
        2. meta for group, which is generated by the founder's private key
        3. meta key, which can verify message sent by this user(or group founder)
        4. visa key, which can encrypt message for the receiver(user)
    """

    @abstractmethod
    def meta(self, identifier: ID) -> Optional[Meta]:
        """
        Get meta for entity ID

        :param identifier: entity ID
        :return:           Meta
        """
        raise NotImplemented

    @abstractmethod
    def document(self, identifier: ID, doc_type: Optional[str]='*') -> Optional[Document]:
        """
        Get document for entity ID

        :param identifier: entity ID
        :param doc_type:   document type
        :return:           Document
        """
        raise NotImplemented


class Entity:
    """Base class of User and Group, ...

        Entity (User/Group)
        ~~~~~~~~~~~~~~~~~~~

            properties:
                identifier - entity ID
                type       - entity type
                meta       - meta for entity ID
                document   - visa for user, or bulletin for group
    """

    def __init__(self, identifier: ID):
        """
        Create Entity with ID

        :param identifier: User/Group ID
        """
        super().__init__()
        self.__identifier: ID = identifier
        self.__delegate: weakref.ReferenceType = None

    def __str__(self):
        clazz = self.__class__.__name__
        return '<%s|%s %s/>' % (clazz, self.type, self.identifier)

    def __eq__(self, other) -> bool:
        if super().__eq__(other):
            return True
        elif isinstance(other, Entity):
            return self.__identifier == other.identifier

    @property
    def delegate(self) -> Optional[EntityDataSource]:
        if self.__delegate is not None:
            return self.__delegate()

    @delegate.setter
    def delegate(self, value: EntityDataSource):
        self.__delegate = weakref.ref(value)

    @property
    def identifier(self) -> ID:
        return self.__identifier

    @property
    def type(self) -> int:
        """ Entity type """
        return self.__identifier.type

    @property
    def meta(self) -> Meta:
        assert isinstance(self.delegate, EntityDataSource), 'entity delegate error: %s' % self.delegate
        return self.delegate.meta(identifier=self.identifier)

    def document(self, doc_type: Optional[str]='*') -> Optional[Document]:
        assert isinstance(self.delegate, EntityDataSource), 'entity delegate error: %s' % self.delegate
        return self.delegate.document(identifier=self.identifier, doc_type=doc_type)



# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Delegates
    ~~~~~~~~~

    Delegates for Transceiver, such as Barrack, KeyStore, FTP, ...
"""

from abc import ABC, abstractmethod
from typing import Optional, List

from mkm.crypto import SymmetricKey
from mkm import ID

from .user import User
from .group import Group


class EntityDelegate(ABC):

    @property
    @abstractmethod
    def local_users(self) -> Optional[List[User]]:
        """
        Get all local users (for decrypting received message)

        :return: users with private key
        """
        raise NotImplemented

    @abstractmethod
    def select_user(self, receiver: ID) -> Optional[User]:
        """
        Select local user for receiver

        :param receiver: user/group ID
        :return: local user
        """
        raise NotImplemented

    @abstractmethod
    def user(self, identifier: ID) -> Optional[User]:
        """
        Create user with ID

        :param identifier: ID object
        :return: User object
        """
        raise NotImplemented

    @abstractmethod
    def group(self, identifier: ID) -> Optional[Group]:
        """
        Create group with ID

        :param identifier: ID object
        :return: Group object
        """
        raise NotImplemented


class CipherKeyDelegate(ABC):

    @abstractmethod
    def cipher_key(self, sender: ID, receiver: ID, generate: bool=False) -> Optional[SymmetricKey]:
        """
        Get cipher key for encrypt message from 'sender' to 'receiver'

        :param sender:   user or contact ID
        :param receiver: contact or user/group ID
        :param generate: generate when key not exists
        :return:         cipher key
        """
        raise NotImplemented

    @abstractmethod
    def cache_cipher_key(self, key: SymmetricKey, sender: ID, receiver: ID):
        """
        Cache cipher key for reusing, with direction (from 'sender' to 'receiver')

        :param key:      cipher key from a contact
        :param sender:   user or contact ID
        :param receiver: contact or user/group ID
        """
        raise NotImplemented



# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import abstractmethod
from typing import Optional, Union

from dkd import ContentType, BaseContent


class Command(BaseContent):
    """
        Command Message Content
        ~~~~~~~~~~~~~~~~~~~~~~~

        data format: {
            type : 0x88,
            sn   : 123,

            command : "...", // command name
            extra   : info   // command parameters
        }
    """

    # -------- command names begin --------
    META = 'meta'
    DOCUMENT = 'document'

    RECEIPT = 'receipt'
    HANDSHAKE = 'handshake'
    LOGIN = 'login'
    # -------- command names end --------

    def __init__(self, cmd: Optional[dict]=None, content_type: Union[ContentType, int]=0, command: Optional[str]=None):
        if cmd is None:
            if content_type is 0:
                content_type = ContentType.COMMAND
            super().__init__(content_type=content_type)
        else:
            super().__init__(content=cmd)
        self.__command = command
        if command is not None:
            self['command'] = command

    @property
    def command(self) -> str:
        if self.__command is None:
            self.__command = command_name(cmd=self.dictionary)
        return self.__command

    #
    #  Factory for creating Command
    #
    class Factory:

        @abstractmethod
        def parse_command(self, cmd: dict):  # -> Optional[Command]:
            """
            Parse map object to command

            :param cmd: command info
            :return: Command
            """
            raise NotImplemented

    __factories = {}  # name -> factory

    @classmethod
    def register(cls, command: str, factory: Factory):
        cls.__factories[command] = factory

    @classmethod
    def factory(cls, command: str) -> Factory:
        return cls.__factories.get(command)


"""
    Implements
    ~~~~~~~~~~
"""


def command_name(cmd: dict) -> str:
    return cmd.get('command')



# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    DIMP - Message Contents & Commands
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Define universal message types as contents and commands
"""

from typing import Optional

from dkd import ContentType, Content, BaseContent
from dkd.content import content_group

from .forward import ForwardContent
from .text import TextContent
from .file import FileContent, ImageContent, AudioContent, VideoContent

from .command import Command, command_name
from .history import HistoryCommand
from .group import GroupCommand, InviteCommand, ExpelCommand, JoinCommand, QuitCommand, QueryCommand, ResetCommand

from .meta import MetaCommand
from .document import DocumentCommand


class ContentFactoryBuilder(Content.Factory):

    def __init__(self, content_class):
        super().__init__()
        self.__class = content_class

    def parse_content(self, content: dict) -> Optional[Content]:
        return self.__class(content=content)


class CommandFactoryBuilder(Command.Factory):

    def __init__(self, command_class):
        super().__init__()
        self.__class = command_class

    def parse_command(self, cmd: dict) -> Optional[Command]:
        return self.__class(cmd=cmd)


class CommandFactory(Content.Factory, Command.Factory):

    def __init__(self):
        super().__init__()

    def parse_content(self, content: dict) -> Optional[Content]:
        name = command_name(cmd=content)
        # get factory by command name
        factory = Command.factory(command=name)
        if factory is None:
            # check for group command
            if content_group(content=content) is not None:
                factory = Command.factory(command='group')
            if factory is None:
                factory = self
        return factory.parse_command(cmd=content)

    def parse_command(self, cmd: dict) -> Optional[Command]:
        return Command(cmd=cmd)


class HistoryCommandFactory(CommandFactory):

    def parse_command(self, cmd: dict) -> Optional[Command]:
        return HistoryCommand(cmd=cmd)


class GroupCommandFactory(HistoryCommandFactory):

    def parse_content(self, content: dict) -> Optional[Content]:
        name = command_name(cmd=content)
        # get factory by command name
        factory = Command.factory(command=name)
        if factory is None:
            factory = self
        return factory.parse_command(cmd=content)

    def parse_command(self, cmd: dict) -> Optional[Command]:
        return GroupCommand(cmd=cmd)


def register_content_factories():
    """ Register core content factories """
    Content.register(content_type=ContentType.FORWARD, factory=ContentFactoryBuilder(content_class=ForwardContent))
    Content.register(content_type=ContentType.TEXT, factory=ContentFactoryBuilder(content_class=TextContent))

    Content.register(content_type=ContentType.FILE, factory=ContentFactoryBuilder(content_class=FileContent))
    Content.register(content_type=ContentType.IMAGE, factory=ContentFactoryBuilder(content_class=ImageContent))
    Content.register(content_type=ContentType.AUDIO, factory=ContentFactoryBuilder(content_class=AudioContent))
    Content.register(content_type=ContentType.VIDEO, factory=ContentFactoryBuilder(content_class=VideoContent))

    # Content.register(content_type=ContentType.PAGE, factory=ContentFactoryBuilder(content_class=PageContent))

    Content.register(content_type=ContentType.COMMAND, factory=CommandFactory())
    Content.register(content_type=ContentType.HISTORY, factory=HistoryCommandFactory())

    Content.register(content_type=0, factory=ContentFactoryBuilder(content_class=BaseContent))


def register_command_factories():
    """ Register core command factories """
    Command.register(command=Command.META, factory=CommandFactoryBuilder(command_class=MetaCommand))

    factory = CommandFactoryBuilder(command_class=DocumentCommand)
    Command.register(command=Command.DOCUMENT, factory=factory)
    Command.register(command='profile', factory=factory)
    Command.register(command='visa', factory=factory)
    Command.register(command='bulletin', factory=factory)

    Command.register(command='group', factory=GroupCommandFactory())
    Command.register(command=GroupCommand.INVITE, factory=CommandFactoryBuilder(command_class=InviteCommand))
    Command.register(command=GroupCommand.EXPEL, factory=CommandFactoryBuilder(command_class=ExpelCommand))
    Command.register(command=GroupCommand.JOIN, factory=CommandFactoryBuilder(command_class=JoinCommand))
    Command.register(command=GroupCommand.QUIT, factory=CommandFactoryBuilder(command_class=QuitCommand))
    Command.register(command=GroupCommand.QUERY, factory=CommandFactoryBuilder(command_class=QueryCommand))
    Command.register(command=GroupCommand.RESET, factory=CommandFactoryBuilder(command_class=ResetCommand))


def register_core_factories():
    register_content_factories()
    register_command_factories()


__all__ = [

    'ForwardContent', 'TextContent',
    'FileContent', 'ImageContent', 'AudioContent', 'VideoContent',

    'Command', 'HistoryCommand', 'GroupCommand',
    'InviteCommand', 'ExpelCommand', 'JoinCommand', 'QuitCommand',
    'QueryCommand', 'ResetCommand',

    'MetaCommand', 'DocumentCommand',

    'ContentFactoryBuilder', 'CommandFactoryBuilder',
    'register_core_factories',
]



# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional, Union

from mkm.crypto import base64_encode, base64_decode, hex_encode, md5
from mkm.crypto import SymmetricKey

from dkd import ContentType, BaseContent


def data_filename(data: bytes, ext: str=None) -> str:
    filename = hex_encode(md5(data))
    if ext is None or len(ext) == 0:
        return filename
    return filename + '.' + ext


class FileContent(BaseContent):
    """
        File Message Content
        ~~~~~~~~~~~~~~~~~~~~

        data format: {
            type : 0x10,
            sn   : 123,

            URL      : "http://", // upload to CDN
            data     : "...",     // if (!URL) base64_encode(fileContent)
            filename : "..."
        }
    """

    def __init__(self, content: Optional[dict]=None, content_type: Union[ContentType, int]=0,
                 filename: Optional[str] = None, data: Optional[bytes]=None):
        if content is None:
            if content_type is 0:
                content_type = ContentType.FILE
            super().__init__(content_type=content_type)
        else:
            super().__init__(content=content)
        self.__attachment = data  # attachment (file data)
        self.__password = None    # symmetric key for decryption
        # set values to inner dictionary
        if filename is not None:
            self['filename'] = filename
        if data is not None:
            self['data'] = base64_encode(data=data)

    # URL
    @property
    def url(self) -> Optional[str]:
        return self.get('URL')

    @url.setter
    def url(self, string: str):
        if string is None:
            self.pop('URL', None)
        else:
            self['URL'] = string

    # file data (it's too big to set in the dictionary)
    @property
    def data(self) -> Optional[bytes]:
        if self.__attachment is None:
            base64 = self.get('data')
            if base64 is not None:
                self.__attachment = base64_decode(base64)
        return self.__attachment

    @data.setter
    def data(self, attachment: bytes):
        self.__attachment = attachment
        if attachment is None:
            self.pop('data', None)
        else:
            self['data'] = base64_encode(attachment)

    # filename
    @property
    def filename(self) -> Optional[str]:
        return self.get('filename')

    @filename.setter
    def filename(self, string: str):
        if string is None:
            self.pop('filename', None)
        else:
            self['filename'] = string

    # password for decrypting the downloaded data from CDN
    @property
    def password(self) -> Optional[SymmetricKey]:
        if self.__password is None:
            self.__password = SymmetricKey.parse(key=self.get('password'))
        return self.__password

    @password.setter
    def password(self, key: SymmetricKey):
        self.__password = key
        if key is None:
            self.pop('password', None)
        else:
            self['password'] = key.dictionary


class ImageContent(FileContent):
    """
        Image Message Content
        ~~~~~~~~~~~~~~~~~~~~~

        data format: {
            type : 0x12,
            sn   : 123,

            URL       : "http://", // upload to CDN
            data      : "...",     // if (!URL) base64_encode(image)
            thumbnail : "...",     // base64_encode(smallImage)
            filename  : "..."
        }
    """

    def __init__(self, content: Optional[dict]=None, filename: Optional[str] = None, data: Optional[bytes]=None,
                 thumbnail: Optional[bytes]=None):
        super().__init__(content, ContentType.IMAGE, filename=filename, data=data)
        self.__thumbnail = thumbnail
        if thumbnail is not None:
            self['thumbnail'] = base64_encode(data=thumbnail)

    # thumbnail of image
    @property
    def thumbnail(self) -> Optional[bytes]:
        if self.__thumbnail is None:
            base64 = self.get('thumbnail')
            if base64 is not None:
                self.__thumbnail = base64_decode(base64)
        return self.__thumbnail

    @thumbnail.setter
    def thumbnail(self, small_image: bytes):
        self.__thumbnail = small_image
        if small_image is None:
            self.pop('thumbnail', None)
        else:
            self['thumbnail'] = base64_encode(small_image)


class AudioContent(FileContent):
    """
        Audio Message Content
        ~~~~~~~~~~~~~~~~~~~~~

        data format: {
            type : 0x14,
            sn   : 123,

            URL      : "http://", // upload to CDN
            data     : "...",     // if (!URL) base64_encode(audio)
            text     : "...",     // Automatic Speech Recognition
            filename : "..."
        }
    """

    def __init__(self, content: Optional[dict]=None, filename: Optional[str] = None, data: Optional[bytes]=None,
                 text: Optional[str]=None):
        super().__init__(content, ContentType.AUDIO, filename=filename, data=data)
        if text is not None:
            self['text'] = text

    # ASR text
    @property
    def text(self) -> Optional[bytes]:
        return self.get('text')

    @text.setter
    def text(self, string: str):
        if string is None:
            self.pop('text', None)
        else:
            self['text'] = string


class VideoContent(FileContent):
    """
        Video Message Content
        ~~~~~~~~~~~~~~~~~~~~~

        data format: {
            type : 0x16,
            sn   : 123,

            URL      : "http://", // upload to CDN
            data     : "...",     // if (!URL) base64_encode(video)
            snapshot : "...",     // base64_encode(smallImage)
            filename : "..."
        }
    """

    def __init__(self, content: Optional[dict]=None, filename: Optional[str] = None, data: Optional[bytes]=None,
                 snapshot: Optional[bytes]=None):
        super().__init__(content, ContentType.VIDEO, filename=filename, data=data)
        self.__snapshot = snapshot
        if snapshot is not None:
            self['snapshot'] = base64_encode(data=snapshot)

    # snapshot of video
    @property
    def snapshot(self) -> Optional[bytes]:
        if self.__snapshot is None:
            base64 = self.get('snapshot')
            if base64 is not None:
                self.__snapshot = base64_decode(base64)
        return self.__snapshot

    @snapshot.setter
    def snapshot(self, small_image: bytes):
        self.__snapshot = small_image
        if small_image is None:
            self.pop('snapshot', None)
        else:
            self['snapshot'] = base64_encode(small_image)



# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Entity Database
    ~~~~~~~~~~~~~~~

    Manage meta for all entities
"""

from abc import abstractmethod
from typing import Optional, List

from mkm.crypto import EncryptKey, VerifyKey
from mkm import NetworkType, ID, ANYONE
from mkm import Document, Visa, Bulletin

from .user import User, UserDataSource
from .group import Group, GroupDataSource

from .delegate import EntityDelegate


def thanos(planet: dict, finger: int) -> int:
    """ Thanos can kill half lives of a world with a snap of the finger """
    keys = planet.keys()
    for key in keys:
        if (++finger & 1) == 1:
            # kill it
            planet.pop(key)
    return finger


class Barrack(EntityDelegate, UserDataSource, GroupDataSource):

    def __init__(self):
        super().__init__()
        # memory caches
        self.__users = {}   # ID -> User
        self.__groups = {}  # ID -> Group

    def reduce_memory(self) -> int:
        """
        Call it when received 'UIApplicationDidReceiveMemoryWarningNotification',
        this will remove 50% of cached objects

        :return: number of survivors
        """
        finger = 0
        finger = thanos(self.__users, finger)
        finger = thanos(self.__groups, finger)
        return finger >> 1

    def cache_user(self, user: User):
        if user.delegate is None:
            user.delegate = self
        self.__users[user.identifier] = user

    def cache_group(self, group: Group):
        if group.delegate is None:
            group.delegate = self
        self.__groups[group.identifier] = group

    @abstractmethod
    def create_user(self, identifier: ID) -> Optional[User]:
        raise NotImplemented

    @abstractmethod
    def create_group(self, identifier: ID) -> Optional[Group]:
        raise NotImplemented

    #
    #   EntityDelegate
    #
    def select_user(self, receiver: ID) -> Optional[User]:
        users = self.local_users
        assert users is not None and len(users) > 0, 'local users should not be empty'
        if receiver.is_broadcast:
            return users[0]
        if receiver.is_group:
            # group message (recipient not designated)
            members = self.members(identifier=receiver)
            if members is None or len(members) == 0:
                # TODO: group not ready, waiting for group info
                return None
            for item in users:
                assert isinstance(item, User), 'local user error: %s' % item
                if item.identifier in members:
                    # DISCUSS: set this item to be current user?
                    return item
        else:
            # 1. personal message
            # 2. split group message
            for item in users:
                assert isinstance(item, User), 'local user error: %s' % item
                if item.identifier == receiver:
                    # DISCUSS: set this item to be current user?
                    return item

    def user(self, identifier: ID) -> Optional[User]:
        # 1. get from user cache
        usr = self.__users.get(identifier)
        if usr is None:
            # 2. create and cache it
            usr = self.create_user(identifier=identifier)
            if usr is not None:
                self.cache_user(user=usr)
        return usr

    def group(self, identifier: ID) -> Optional[Group]:
        # 1. get from group cache
        grp = self.__groups.get(identifier)
        if grp is None:
            # 2. create and cache it
            grp = self.create_group(identifier=identifier)
            if grp is not None:
                self.cache_group(group=grp)
        return grp

    #
    #   UserDataSource
    #
    def __visa_key(self, identifier: ID) -> Optional[EncryptKey]:
        visa = self.document(identifier=identifier, doc_type=Document.VISA)
        if isinstance(visa, Visa):
            if visa.valid:
                return visa.key

    def __meta_key(self, identifier: ID) -> Optional[VerifyKey]:
        meta = self.meta(identifier=identifier)
        if meta is not None:
            return meta.key

    def public_key_for_encryption(self, identifier: ID) -> Optional[EncryptKey]:
        # 1. get key from visa
        key = self.__visa_key(identifier=identifier)
        if key is not None:
            # if visa.key exists, use it for encryption
            return key
        # 2. get key from meta
        key = self.__meta_key(identifier=identifier)
        if isinstance(key, EncryptKey):
            # if visa.key not exists and meta.key is encrypt key,
            # use it for encryption
            return key

    def public_keys_for_verification(self, identifier: ID) -> Optional[List[VerifyKey]]:
        keys = []
        # 1. get key from visa
        key = self.__visa_key(identifier=identifier)
        if isinstance(key, VerifyKey):
            # the sender may use communication key to sign message.data,
            # so try to verify it with visa.key here
            keys.append(key)
        # 2. get key from meta
        key = self.__meta_key(identifier=identifier)
        if key is not None:
            # the sender may use identity key to sign message.data,
            # try to verify it with meta.key
            keys.append(key)
        assert len(keys) > 0, 'failed to get verify key for user: %s' % identifier
        return keys

    #
    #   GroupDataSource
    #
    def founder(self, identifier: ID) -> Optional[ID]:
        # check for broadcast
        if identifier.is_broadcast:
            # founder of broadcast group
            name = identifier.name
            if name is None:
                length = 0
            else:
                length = len(name)
            if length == 0 or (length == 8 and name == 'everyone'):
                # Consensus: the founder of group 'everyone@everywhere'
                #            'Albert Moky'
                founder = 'moky@anywhere'
            else:
                # DISCUSS: who should be the founder of group 'xxx@everywhere'?
                #          'anyone@anywhere', or 'xxx.founder@anywhere'
                founder = name + '.founder@anywhere'
            return ID.parse(identifier=founder)
        # check group meta
        g_meta = self.meta(identifier=identifier)
        if g_meta is None:
            # FIXME: when group profile was arrived but the meta still on the way,
            #        here will cause founder not found
            return None
        # check each member's public key with group meta
        members = self.members(identifier=identifier)
        if members is not None:
            for item in members:
                u_meta = self.meta(identifier=item)
                if u_meta is None:
                    # failed to get member's meta
                    continue
                if g_meta.match_key(key=u_meta.key):
                    # if the member's public key matches with the group's meta,
                    # it means this meta was generated by the member's private key
                    return item
        # TODO: load founder from database

    def owner(self, identifier: ID) -> Optional[ID]:
        # check for broadcast
        if identifier.is_broadcast:
            # owner of broadcast group
            name = identifier.name
            if name is None:
                length = 0
            else:
                length = len(name)
            if length == 0 or (length == 8 and name == 'everyone'):
                # Consensus: the owner of group 'everyone@everywhere'
                #            'anyone@anywhere'
                owner = ANYONE
            else:
                # DISCUSS: who should be the owner of group 'xxx@everywhere'?
                #          'anyone@anywhere', or 'xxx.owner@anywhere'
                owner = name + '.owner@anywhere'
            return ID.parse(identifier=owner)
        # check group type
        if identifier.type == NetworkType.POLYLOGUE:
            # Polylogue's owner is its founder
            return self.founder(identifier=identifier)
        # TODO: load owner from database

    def members(self, identifier: ID) -> Optional[List[ID]]:
        # check for broadcast
        if identifier.is_broadcast:
            # members of broadcast group
            name = identifier.name
            if name is None:
                length = 0
            else:
                length = len(name)
            if length == 0 or (length == 8 and name == 'everyone'):
                # Consensus: the member of group 'everyone@everywhere'
                #            'anyone@anywhere'
                member = ANYONE
                owner = ANYONE
            else:
                # DISCUSS: who should be the member of group 'xxx@everywhere'?
                #          'anyone@anywhere', or 'xxx.member@anywhere'
                member = name + '.member@anywhere'
                owner = name + '.owner@anywhere'
            member = ID.parse(identifier=member)
            owner = ID.parse(identifier=owner)
            if member == owner:
                return [owner]
            else:
                return [owner, member]

    def assistants(self, identifier: ID) -> Optional[List[ID]]:
        doc = self.document(identifier=identifier, doc_type=Document.BULLETIN)
        if isinstance(doc, Bulletin):
            if doc.valid:
                return doc.assistants
        # TODO: get group bots from SP configuration



# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from mkm.crypto import *
from mkm import *
from dkd import *

from .protocol import *

from .entity import Entity, EntityDataSource
from .user import User, UserDataSource
from .group import Group, GroupDataSource

from .delegate import EntityDelegate, CipherKeyDelegate

from .barrack import Barrack
from .transceiver import Transceiver
from .packer import Packer
from .processor import Processor

name = "DIMP"

__author__ = 'Albert Moky'

__all__ = [

    #
    #   Crypto
    #
    'DataCoder', 'Base64', 'Base58', 'Hex',
    'base64_encode', 'base64_decode', 'base58_encode', 'base58_decode', 'hex_encode', 'hex_decode',
    'DataParser', 'JSON', 'UTF8',
    'json_encode', 'json_decode', 'utf8_encode', 'utf8_decode',

    'DataDigester', 'MD5', 'SHA1', 'SHA256', 'KECCAK256', 'RIPEMD160',
    'md5', 'sha1', 'sha256', 'keccak256', 'ripemd160',

    'Wrapper', 'Array', 'Map', 'Dictionary', 'String',

    'CryptographyKey',
    'SymmetricKey', 'EncryptKey', 'DecryptKey',
    'AsymmetricKey', 'SignKey', 'VerifyKey',
    'PublicKey', 'PrivateKey',

    #
    #   MingKeMing
    #
    'NetworkType', 'MetaType',
    'Address', 'AddressFactory', 'ANYWHERE', 'EVERYWHERE',
    'ID', 'ANYONE', 'EVERYONE', 'FOUNDER',
    'Meta', 'BaseMeta',
    'Document', 'BaseDocument',
    'Visa', 'BaseVisa', 'Bulletin', 'BaseBulletin',

    #
    #   DaoKeDao
    #
    'ContentType',
    'Content', 'BaseContent',
    'Envelope',
    'Message',
    'InstantMessage', 'SecureMessage', 'ReliableMessage',
    'MessageDelegate',
    'InstantMessageDelegate', 'SecureMessageDelegate', 'ReliableMessageDelegate',

    #
    #   Protocol
    #
    'ForwardContent', 'TextContent',
    'FileContent', 'ImageContent', 'AudioContent', 'VideoContent',

    'Command', 'HistoryCommand', 'GroupCommand',
    'InviteCommand', 'ExpelCommand', 'JoinCommand', 'QuitCommand',
    'QueryCommand', 'ResetCommand',

    'MetaCommand', 'DocumentCommand',

    #
    #   Core
    #
    'Entity', 'User', 'Group',
    'EntityDataSource', 'UserDataSource', 'GroupDataSource',

    'EntityDelegate', 'CipherKeyDelegate',
    'Barrack', 'Transceiver', 'Packer', 'Processor',
]



# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import weakref
from abc import ABC
from typing import Optional

from dkd import Envelope, InstantMessage, SecureMessage, ReliableMessage

from .delegate import EntityDelegate
from .transceiver import Transceiver


class Processor(Transceiver.Processor, ABC):

    def __init__(self, transceiver: Transceiver):
        super().__init__()
        self.__transceiver = weakref.ref(transceiver)

    @property
    def transceiver(self) -> Transceiver:
        return self.__transceiver()

    @property
    def barrack(self) -> EntityDelegate:
        return self.transceiver.barrack

    @property
    def packer(self) -> Transceiver.Packer:
        return self.transceiver.packer

    #
    #  Processing Message
    #

    def process_package(self, data: bytes) -> Optional[bytes]:
        # 1. deserialize message
        msg = self.packer.deserialize_message(data=data)
        if msg is None:
            # no valid message received
            return None
        # 2. process message
        msg = self.transceiver.process_reliable_message(msg=msg)
        if msg is None:
            # nothing to respond
            return None
        # 3. serialize message
        return self.packer.serialize_message(msg=msg)

    def process_reliable_message(self, msg: ReliableMessage) -> Optional[ReliableMessage]:
        # TODO: override to check broadcast message before calling it
        # 1. verify message
        s_msg = self.packer.verify_message(msg=msg)
        if s_msg is None:
            # waiting for sender's meta if not exists
            return None
        # 2. process message
        s_msg = self.transceiver.process_secure_message(msg=s_msg, r_msg=msg)
        if s_msg is None:
            # nothing to respond
            return None
        # 3. sign message
        return self.packer.sign_message(msg=s_msg)
        # TODO: override to deliver to the receiver when catch exception "receiver error ..."

    def process_secure_message(self, msg: SecureMessage, r_msg: ReliableMessage) -> Optional[SecureMessage]:
        # 1. decrypt message
        i_msg = self.packer.decrypt_message(msg=msg)
        if i_msg is None:
            # cannot decrypt this message, not for you?
            # delivering message to other receiver?
            return None
        # 2. process message
        i_msg = self.transceiver.process_instant_message(msg=i_msg, r_msg=r_msg)
        if i_msg is None:
            # nothing to respond
            return None
        # 3. encrypt message
        return self.packer.encrypt_message(msg=i_msg)

    def process_instant_message(self, msg: InstantMessage, r_msg: ReliableMessage) -> Optional[InstantMessage]:
        # process content from sender
        res = self.transceiver.process_content(content=msg.content, r_msg=r_msg)
        if res is None:
            # nothing to respond
            return None
        user = self.barrack.select_user(receiver=msg.receiver)
        if user is None:
            raise AssertionError('receiver error: %s' % msg.receiver)
        # pack message
        env = Envelope.create(sender=user.identifier, receiver=msg.sender)
        return InstantMessage.create(head=env, body=res)



# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional

from dkd import ContentType

from .command import Command


class HistoryCommand(Command):
    """
        History Command
        ~~~~~~~~~~~~~~~

        data format: {
            type : 0x89,
            sn   : 123,

            command : "...", // command name
            time    : 0,     // command timestamp
            extra   : info   // command parameters
        }
    """

    # -------- command names begin --------
    # account
    REGISTER = "register"
    SUICIDE = "suicide"

    # group: founder/owner
    FOUND = "found"
    ABDICATE = "abdicate"
    # group: member
    INVITE = "invite"
    EXPEL = "expel"
    JOIN = "join"
    QUIT = "quit"
    RESET = "reset"
    QUERY = "query"  # (not history command)
    # group: administrator/assistant
    HIRE = "hire"
    FIRE = "fire"
    RESIGN = "resign"
    # -------- command names end --------

    def __init__(self, cmd: Optional[dict]=None, command: Optional[str]=None):
        super().__init__(cmd=cmd, content_type=ContentType.HISTORY, command=command)



# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import weakref
from typing import Optional

from mkm.crypto import json_encode, json_decode

from mkm import ID
from dkd import Content
from dkd import InstantMessage, SecureMessage, ReliableMessage

from .protocol import Command

from .delegate import EntityDelegate, CipherKeyDelegate
from .transceiver import Transceiver


class Packer(Transceiver.Packer):

    def __init__(self, transceiver: Transceiver):
        super().__init__()
        self.__transceiver = weakref.ref(transceiver)

    @property
    def transceiver(self) -> Transceiver:
        return self.__transceiver()

    @property
    def barrack(self) -> EntityDelegate:
        return self.transceiver.barrack

    @property
    def key_cache(self) -> CipherKeyDelegate:
        return self.transceiver.key_cache

    def overt_group(self, content: Content) -> Optional[ID]:
        group = content.group
        if group is not None:
            if group.is_broadcast:
                # broadcast message is always overt
                return group
            if isinstance(content, Command):
                # group command should be sent to each member directly, so
                # don't expose group ID
                return None
            return group

    #
    #   InstantMessage -> SecureMessage -> ReliableMessage -> Data
    #

    def encrypt_message(self, msg: InstantMessage) -> Optional[SecureMessage]:
        # check message delegate
        if msg.delegate is None:
            msg.delegate = self.transceiver

        sender = msg.sender
        receiver = msg.receiver
        # if 'group' exists and the 'receiver' is a group ID,
        # they must be equal

        # NOTICE: while sending group message, don't split it before encrypting.
        #         this means you could set group ID into message content, but
        #         keep the "receiver" to be the group ID;
        #         after encrypted (and signed), you could split the message
        #         with group members before sending out, or just send it directly
        #         to the group assistant to let it split messages for you!
        #    BUT,
        #         if you don't want to share the symmetric key with other members,
        #         you could split it (set group ID into message content and
        #         set contact ID to the "receiver") before encrypting, this usually
        #         for sending group command to assistant robot, which should not
        #         share the symmetric key (group msg key) with other members.

        # 1. get symmetric key
        group = self.overt_group(content=msg.content)
        if group is None:
            # personal message or (group) command
            password = self.key_cache.cipher_key(sender=sender, receiver=receiver, generate=True)
            assert password is not None, 'failed to get msg key: %s -> %s' % (sender, receiver)
        else:
            # group message (excludes group command)
            password = self.key_cache.cipher_key(sender=sender, receiver=group, generate=True)
            assert password is not None, 'failed to get group msg key: %s -> %s' % (sender, group)

        # 2. encrypt 'content' to 'data' for receiver/group members
        if receiver.is_group:
            # group message
            grp = self.barrack.group(identifier=receiver)
            if grp is None:
                # group not ready
                # TODO: suspend this message for waiting group's meta
                return None
            members = grp.members
            if members is None or len(members) == 0:
                # group members not found
                # TODO: suspend this message for waiting group's membership
                return None
            s_msg = msg.encrypt(password=password, members=grp.members)
        else:
            # personal message (or split group message)
            s_msg = msg.encrypt(password=password)
        if s_msg is None:
            # public key for encryption not found
            # TODO: suspend this message for waiting receiver's meta
            return None

        # overt group ID
        if group is not None and receiver != group:
            # NOTICE: this help the receiver knows the group ID
            #         when the group message separated to multi-messages,
            #         if don't want the others know you are the group members,
            #         remove it.
            s_msg.envelope.group = group

        # NOTICE: copy content type to envelope
        #         this help the intermediate nodes to recognize message type
        s_msg.envelope.type = msg.content.type

        # OK
        return s_msg

    def sign_message(self, msg: SecureMessage) -> ReliableMessage:
        # check message delegate
        if msg.delegate is None:
            msg.delegate = self.transceiver
        assert msg.data is not None, 'message data cannot be empty: %s' % msg
        # sign 'data' by sender
        return msg.sign()

    def serialize_message(self, msg: ReliableMessage) -> bytes:
        return json_encode(o=msg.dictionary)

    #
    #   Data -> ReliableMessage -> SecureMessage -> InstantMessage
    #

    def deserialize_message(self, data: bytes) -> Optional[ReliableMessage]:
        dictionary = json_decode(data=data)
        # TODO: translate short keys
        #       'S' -> 'sender'
        #       'R' -> 'receiver'
        #       'W' -> 'time'
        #       'T' -> 'type'
        #       'G' -> 'group'
        #       ------------------
        #       'D' -> 'data'
        #       'V' -> 'signature'
        #       'K' -> 'key'
        #       ------------------
        #       'M' -> 'meta'
        return ReliableMessage.parse(msg=dictionary)

    def verify_message(self, msg: ReliableMessage) -> Optional[SecureMessage]:
        # check message delegate
        if msg.delegate is None:
            msg.delegate = self.transceiver
        #
        # TODO: check [Meta Protocol]
        #       make sure the sender's meta exists
        #       (do in by application)
        #
        assert msg.signature is not None, 'message signature cannot be empty: %s' % msg
        # verify 'data' with 'signature'
        return msg.verify()

    def decrypt_message(self, msg: SecureMessage) -> Optional[InstantMessage]:
        # check message delegate
        if msg.delegate is None:
            msg.delegate = self.transceiver
        #
        # NOTICE: make sure the receiver is YOU!
        #         which means the receiver's private key exists;
        #         if the receiver is a group ID, split it first
        #

        assert msg.data is not None, 'message data cannot be empty: %s' % msg
        # decrypt 'data' to 'content'
        return msg.decrypt()
        # TODO: check top-secret message
        #       (do it by application)



# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import abstractmethod
from typing import Optional, List

from mkm.crypto import EncryptKey, DecryptKey, SignKey, VerifyKey
from mkm import ID, Visa, Document

from .entity import Entity, EntityDataSource


class UserDataSource(EntityDataSource):
    """This interface is for getting information for user

        User Data Source
        ~~~~~~~~~~~~~~~~

        (Encryption/decryption)
        1. public key for encryption
           if visa.key not exists, means it is the same key with meta.key
        2. private keys for decryption
           the private keys paired with [visa.key, meta.key]

        (Signature/Verification)
        3. private key for signature
           the private key paired with visa.key or meta.key
        4. public keys for verification
           [visa.key, meta.key]

        (Visa Document)
        5. private key for visa signature
           the private key pared with meta.key
        6. public key for visa verification
           meta.key only
    """

    @abstractmethod
    def contacts(self, identifier: ID) -> Optional[List[ID]]:
        """
        Get user's contacts list

        :param identifier: user ID
        :return: contact ID list
        """
        pass

    @abstractmethod
    def public_key_for_encryption(self, identifier: ID) -> Optional[EncryptKey]:
        """
        Get user's public key for encryption
        (visa.key or meta.key)

        :param identifier: user ID
        :return: public key
        """
        pass

    @abstractmethod
    def public_keys_for_verification(self, identifier: ID) -> Optional[List[VerifyKey]]:
        """
        Get user's public keys for verification
        [visa.key, meta.key]

        :param identifier: user ID
        :return: public keys
        """
        pass

    @abstractmethod
    def private_keys_for_decryption(self, identifier: ID) -> Optional[List[DecryptKey]]:
        """
        Get user's private keys for decryption
        (which paired with [visa.key, meta.key])

        :param identifier: user ID
        :return: private keys
        """
        raise NotImplemented

    @abstractmethod
    def private_key_for_signature(self, identifier: ID) -> Optional[SignKey]:
        """
        Get user's private key for signature
        (which paired with visa.key or meta.key)

        :param identifier: user ID
        :return: private key
        """
        raise NotImplemented

    @abstractmethod
    def private_key_for_visa_signature(self, identifier: ID) -> Optional[SignKey]:
        """
        Get user's private key for signing visa

        :param identifier: user ID
        :return: private key
        """
        raise NotImplemented


class User(Entity):
    """This class is for creating user

        User for communication
        ~~~~~~~~~~~~~~~~~~~~~~

        functions:
            (User)
            1. verify(data, signature) - verify (encrypted content) data and signature
            2. encrypt(data)           - encrypt (symmetric key) data
            (LocalUser)
            3. sign(data)    - calculate signature of (encrypted content) data
            4. decrypt(data) - decrypt (symmetric key) data
    """

    @Entity.delegate.getter
    def delegate(self) -> Optional[UserDataSource]:
        return super().delegate

    # @delegate.setter
    # def delegate(self, value: UserDataSource):
    #     super(User, User).delegate.__set__(self, value)

    @property
    def visa(self) -> Optional[Visa]:
        doc = self.document(doc_type=Document.VISA)
        if isinstance(doc, Visa):
            return doc

    @property
    def contacts(self) -> List[ID]:
        """
        Get all contacts of the user

        :return: contacts list
        """
        assert isinstance(self.delegate, UserDataSource), 'user data source error: %s' % self.delegate
        return self.delegate.contacts(identifier=self.identifier)

    def verify(self, data: bytes, signature: bytes) -> bool:
        """
        Verify data and signature with user's public keys

        :param data:
        :param signature:
        :return:
        """
        assert isinstance(self.delegate, UserDataSource), 'user data source error: %s' % self.delegate
        # NOTICE: I suggest using the private key paired with meta.key to sign message,
        #         so here should return the meta.key
        keys = self.delegate.public_keys_for_verification(identifier=self.identifier)
        for key in keys:
            if key.verify(data=data, signature=signature):
                # matched!
                return True

    def encrypt(self, data: bytes) -> bytes:
        """Encrypt data, try visa.key first, if not found, use meta.key

        :param data: plaintext
        :return: encrypted data
        """
        assert isinstance(self.delegate, UserDataSource), 'user data source error: %s' % self.delegate
        # NOTICE: meta.key will never changed, so use visa.key to encrypt message
        #         is the better way
        key = self.delegate.public_key_for_encryption(identifier=self.identifier)
        assert key is not None, 'failed to get encrypt key for user: %s' % self.identifier
        return key.encrypt(data=data)

    #
    #   interfaces for local user
    #
    def sign(self, data: bytes) -> bytes:
        """
        Sign data with user's private key

        :param data: message data
        :return: signature
        """
        assert isinstance(self.delegate, UserDataSource), 'user data source error: %s' % self.delegate
        # NOTICE: I suggest use the private key which paired to meta.key
        #         to sign message
        key = self.delegate.private_key_for_signature(identifier=self.identifier)
        assert key is not None, 'failed to get sign key for user: %s' % self.identifier
        return key.sign(data=data)

    def decrypt(self, data: bytes) -> Optional[bytes]:
        """
        Decrypt data with user's private key(s)

        :param data: encrypted data
        :return: plaintext
        """
        assert isinstance(self.delegate, UserDataSource), 'user data source error: %s' % self.delegate
        # NOTICE: if you provide a public key in visa document for encryption,
        #         here you should return the private key paired with visa.key
        keys = self.delegate.private_keys_for_decryption(identifier=self.identifier)
        assert keys is not None and len(keys) > 0, 'failed to get decrypt keys: %s' % self.identifier
        for key in keys:
            try:
                # try decrypting it with each private key
                plaintext = key.decrypt(data=data)
                if plaintext is not None:
                    # OK!
                    return plaintext
            except ValueError:
                # this key not match, try next one
                continue

    #
    #   Interfaces for Visa
    #
    def sign_visa(self, visa: Visa) -> Optional[Visa]:
        if self.identifier == visa.identifier:
            assert isinstance(self.delegate, UserDataSource), 'user data source error: %s' % self.delegate
            # NOTICE: only sign visa with the private key paired with your meta.key
            key = self.delegate.private_key_for_visa_signature(identifier=self.identifier)
            assert key is not None, 'failed to get sign key for visa: %s' % self.identifier
            visa.sign(private_key=key)
            return visa

    def verify_visa(self, visa: Visa) -> bool:
        if self.identifier == visa.identifier:
            # NOTICE: only verify visa with meta.key
            #         (if meta not exists, user won't be created)
            return visa.verify(public_key=self.meta.key)



# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional

from dkd import ContentType, BaseContent
from dkd import ReliableMessage


class ForwardContent(BaseContent):
    """
        Top-Secret Message Content
        ~~~~~~~~~~~~~~~~~~~~~~~~~~

        data format: {
            type : 0xFF,
            sn   : 456,

            forward : {...}  // reliable (secure + certified) message
        }
    """

    def __init__(self, content: Optional[dict]=None, message: Optional[ReliableMessage]=None):
        if content is None:
            super().__init__(content_type=ContentType.FORWARD)
        else:
            super().__init__(content=content)
        self.__forward = None
        if message is not None:
            self.message = message

    #
    #   forward (top-secret message)
    #
    @property
    def message(self) -> ReliableMessage:
        if self.__forward is None:
            self.__forward = ReliableMessage.parse(msg=self.get('forward'))
        return self.__forward

    @message.setter
    def message(self, secret: ReliableMessage):
        self.__forward = secret
        if secret is None:
            self.pop('forward', None)
        else:
            self['forward'] = secret.dictionary



# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Group Command Protocol
    ~~~~~~~~~~~~~~~~~~~~~~

    1. invite member
    2. expel member
    3. member quit
"""

from typing import Optional, List

from mkm import ID

from .history import HistoryCommand


class GroupCommand(HistoryCommand):
    """
        Group Command
        ~~~~~~~~~~~~~

        data format: {
            type : 0x89,
            sn   : 123,

            command : "invite",         // "expel", "quit"
            time    : 0,                // timestamp
            group   : "{GROUP_ID}",     // group ID
            member  : "{MEMBER_ID}",    // member ID
            members : ["{MEMBER_ID}",], // member ID list
        }
    """

    def __init__(self, cmd: Optional[dict]=None, command: Optional[str]=None, group: Optional[ID]=None,
                 member: Optional[ID]=None, members: Optional[List[ID]]=None):
        super().__init__(cmd=cmd, command=command)
        if group is not None:
            self.group = group
        if member is not None:
            self.member = member
        elif members is not None:
            self.members = members

    #
    #   group (group ID already fetched in Content)
    #

    #
    #   member/members
    #
    @property
    def member(self) -> Optional[ID]:
        return ID.parse(identifier=self.get('member'))

    @member.setter
    def member(self, value: ID):
        if value is None:
            self.pop('member', None)
        else:
            self['member'] = str(value)

    @property
    def members(self) -> Optional[List[ID]]:
        array = self.get('members')
        if array is not None:
            # convert all items to ID objects
            return ID.convert(members=array)

    @members.setter
    def members(self, value: List[ID]):
        if value is None:
            self.pop('members', None)
        else:
            self['members'] = ID.revert(members=value)

    @classmethod
    def invite(cls, group: ID, member: Optional[ID]=None, members: Optional[List[ID]]=None):
        return InviteCommand(group=group, member=member, members=members)

    @classmethod
    def expel(cls, group: ID, member: Optional[ID]=None, members: Optional[List[ID]]=None):
        return ExpelCommand(group=group, member=member, members=members)

    @classmethod
    def join(cls, group: ID):
        return JoinCommand(group=group)

    @classmethod
    def quit(cls, group: ID):
        return QuitCommand(group=group)

    @classmethod
    def query(cls, group: ID):
        return QueryCommand(group=group)

    @classmethod
    def reset(cls, group: ID, members: Optional[List[ID]]=None):
        return ResetCommand(group=group, members=members)


class InviteCommand(GroupCommand):

    def __init__(self, cmd: Optional[dict]=None, group: Optional[ID]=None,
                 member: Optional[ID]=None, members: Optional[List[ID]]=None):
        """
        Create invite group member command

        :param cmd:     command info
        :param group:   group ID
        :param member:  member ID
        :param members: member list
        :return: InviteCommand object
        """
        super().__init__(cmd, GroupCommand.INVITE, group=group, member=member, members=members)


class ExpelCommand(GroupCommand):

    def __init__(self, cmd: Optional[dict]=None, group: Optional[ID]=None,
                 member: Optional[ID]=None, members: Optional[List[ID]]=None):
        """
        Create expel group member command

        :param cmd:     command info
        :param group:   group ID
        :param member:  member ID
        :param members: member list
        :return: ExpelCommand object
        """
        super().__init__(cmd, GroupCommand.EXPEL, group=group, member=member, members=members)


class JoinCommand(GroupCommand):

    def __init__(self, cmd: Optional[dict]=None, group: Optional[ID]=None):
        """
        Create join group command

        :param cmd:     command info
        :param group:   group ID
        :return: JoinCommand object
        """
        super().__init__(cmd, GroupCommand.JOIN, group=group)


class QuitCommand(GroupCommand):

    def __init__(self, cmd: Optional[dict]=None, group: Optional[ID]=None):
        """
        Create member quit command

        :param cmd:     command info
        :param group:   group ID
        :return: QuitCommand object
        """
        super().__init__(cmd, GroupCommand.QUIT, group=group)


class QueryCommand(GroupCommand):

    def __init__(self, cmd: Optional[dict]=None, group: Optional[ID]=None):
        """
        Create query group members command (not history command)

        :param cmd:     command info
        :param group:   group ID
        :return: QueryCommand object
        """
        super().__init__(cmd, GroupCommand.QUERY, group=group)


class ResetCommand(GroupCommand):

    def __init__(self, cmd: Optional[dict]=None, group: Optional[ID]=None, members: Optional[List[ID]]=None):
        """
        Create reset group members command

        :param cmd:     command info
        :param group:   group ID
        :param members: member list
        :return: ResetCommand object
        """
        super().__init__(cmd, GroupCommand.RESET, group=group, members=members)



# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Transceiver
    ~~~~~~~~~~~

    Delegate to process message transforming
"""

import weakref
from abc import abstractmethod
from typing import Optional

from mkm.crypto import base64_encode, base64_decode, utf8_encode, utf8_decode, json_encode, json_decode
from mkm.crypto import SymmetricKey
from mkm import ID

from dkd import Message
from dkd import InstantMessage, SecureMessage, ReliableMessage
from dkd import InstantMessageDelegate, ReliableMessageDelegate
from dkd import Content

from .delegate import EntityDelegate, CipherKeyDelegate


class Transceiver(InstantMessageDelegate, ReliableMessageDelegate):

    def __init__(self):
        super().__init__()
        self.__barrack: weakref.ReferenceType = None
        self.__key_cache: weakref.ReferenceType = None
        self.__packer: weakref.ReferenceType = None
        self.__processor: weakref.ReferenceType = None

    @property
    def barrack(self) -> EntityDelegate:
        """ Barrack (EntityDelegate) """
        if self.__barrack is not None:
            return self.__barrack()

    @barrack.setter
    def barrack(self, delegate: EntityDelegate):
        self.__barrack = weakref.ref(delegate)

    @property
    def key_cache(self) -> CipherKeyDelegate:
        """ KeyCache (CipherKeyDelegate) """
        if self.__key_cache is not None:
            return self.__key_cache()

    @key_cache.setter
    def key_cache(self, delegate: CipherKeyDelegate):
        self.__key_cache = weakref.ref(delegate)

    #
    #   Message Packer
    #
    class Packer:

        @abstractmethod
        def overt_group(self, content: Content) -> Optional[ID]:
            """
            Get group ID which should be exposed to public network

            :param content: message content
            :return: exposed group ID
            """
            raise NotImplemented

        @abstractmethod
        def encrypt_message(self, msg: InstantMessage) -> Optional[SecureMessage]:
            """
            Encrypt message content

            :param msg: plain message
            :return: encrypted message
            """
            raise NotImplemented

        @abstractmethod
        def sign_message(self, msg: SecureMessage) -> ReliableMessage:
            """
            Sign content data

            :param msg: encrypted message
            :return: network message
            """
            raise NotImplemented

        @abstractmethod
        def serialize_message(self, msg: ReliableMessage) -> bytes:
            """
            Serialize network message

            :param msg: network message
            :return: data package
            """
            raise NotImplemented

        @abstractmethod
        def deserialize_message(self, data: bytes) -> Optional[ReliableMessage]:
            """
            Deserialize network message

            :param data: data package
            :return: network message
            """
            raise NotImplemented

        @abstractmethod
        def verify_message(self, msg: ReliableMessage) -> Optional[SecureMessage]:
            """
            Verify encrypted content data

            :param msg: network message
            :return: encrypted message
            """
            raise NotImplemented

        @abstractmethod
        def decrypt_message(self, msg: SecureMessage) -> Optional[InstantMessage]:
            """
            Decrypt message content

            :param msg: encrypted message
            :return: plain message
            """
            raise NotImplemented

    @property
    def packer(self) -> Packer:
        """ Message Packer """
        if self.__packer is not None:
            return self.__packer()

    @packer.setter
    def packer(self, delegate: Packer):
        self.__packer = weakref.ref(delegate)

    #
    #   Interfaces for Packing Message
    #
    def overt_group(self, content: Content) -> Optional[ID]:
        return self.packer.overt_group(content=content)

    def encrypt_message(self, msg: InstantMessage) -> Optional[SecureMessage]:
        return self.packer.encrypt_message(msg=msg)

    def sign_message(self, msg: SecureMessage) -> ReliableMessage:
        return self.packer.sign_message(msg=msg)

    def serialize_message(self, msg: ReliableMessage) -> bytes:
        return self.packer.serialize_message(msg=msg)

    def deserialize_message(self, data: bytes) -> Optional[ReliableMessage]:
        return self.packer.deserialize_message(data=data)

    def verify_message(self, msg: ReliableMessage) -> Optional[SecureMessage]:
        return self.packer.verify_message(msg=msg)

    def decrypt_message(self, msg: SecureMessage) -> Optional[InstantMessage]:
        return self.packer.decrypt_message(msg=msg)

    #
    #   Message Processor
    #
    class Processor:

        @abstractmethod
        def process_package(self, data: bytes) -> Optional[bytes]:
            """
            Process data package

            :param data: data to be processed
            :return: response data
            """
            raise NotImplemented

        @abstractmethod
        def process_reliable_message(self, msg: ReliableMessage) -> Optional[ReliableMessage]:
            """
            Process network message

            :param msg: message to be processed
            :return: response message
            """
            raise NotImplemented

        @abstractmethod
        def process_secure_message(self, msg: SecureMessage, r_msg: ReliableMessage) -> Optional[SecureMessage]:
            """
            Process encrypted message

            :param msg:   message to be processed
            :param r_msg: message received
            :return: response message
            """
            raise NotImplemented

        @abstractmethod
        def process_instant_message(self, msg: InstantMessage, r_msg: ReliableMessage) -> Optional[InstantMessage]:
            """
            Process plain message

            :param msg:   message to be processed
            :param r_msg: message received
            :return: response message
            """
            raise NotImplemented

        @abstractmethod
        def process_content(self, content: Content, r_msg: ReliableMessage) -> Optional[Content]:
            """
            Process message content

            :param content: content to be processed
            :param r_msg: message received
            :return: response content
            """
            raise NotImplemented

    @property
    def processor(self) -> Processor:
        """ Message Processor """
        if self.__processor is not None:
            return self.__processor()

    @processor.setter
    def processor(self, delegate: Processor):
        self.__processor = weakref.ref(delegate)

    #
    #   Interfaces for Processing Message
    #
    def process_package(self, data: bytes) -> Optional[bytes]:
        return self.processor.process_package(data=data)

    def process_reliable_message(self, msg: ReliableMessage) -> Optional[ReliableMessage]:
        return self.processor.process_reliable_message(msg=msg)

    def process_secure_message(self, msg: SecureMessage, r_msg: ReliableMessage) -> Optional[SecureMessage]:
        return self.processor.process_secure_message(msg=msg, r_msg=r_msg)

    def process_instant_message(self, msg: InstantMessage, r_msg: ReliableMessage) -> Optional[InstantMessage]:
        return self.processor.process_instant_message(msg=msg, r_msg=r_msg)

    def process_content(self, content: Content, r_msg: ReliableMessage) -> Optional[Content]:
        return self.processor.process_content(content=content, r_msg=r_msg)

    #
    #   InstantMessageDelegate
    #

    def serialize_content(self, content: Content, key: SymmetricKey, msg: InstantMessage) -> bytes:
        # NOTICE: check attachment for File/Image/Audio/Video message content
        #         before serialize content, this job should be do in subclass
        return json_encode(o=content.dictionary)

    def encrypt_content(self, data: bytes, key: SymmetricKey, msg: InstantMessage) -> bytes:
        return key.encrypt(data=data)

    def encode_data(self, data: bytes, msg: InstantMessage) -> str:
        if is_broadcast(msg=msg):
            # broadcast message content will not be encrypted (just encoded to JsON),
            # so no need to encode to Base64 here
            return utf8_decode(data=data)
        return base64_encode(data=data)

    def serialize_key(self, key: SymmetricKey, msg: InstantMessage) -> Optional[bytes]:
        if is_broadcast(msg=msg):
            # broadcast message has no key
            return None
        return json_encode(o=key.dictionary)

    def encrypt_key(self, data: bytes, receiver: ID, msg: InstantMessage) -> Optional[bytes]:
        assert not is_broadcast(msg=msg), 'broadcast message has no key: %s' % msg
        # TODO: make sure the receiver's public key exists
        contact = self.barrack.user(identifier=receiver)
        assert contact is not None, 'failed to encrypt for receiver: %s' % receiver
        # encrypt with receiver's public key
        return contact.encrypt(data=data)

    def encode_key(self, data: bytes, msg: InstantMessage) -> str:
        assert not is_broadcast(msg=msg), 'broadcast message has no key: %s' % msg
        return base64_encode(data=data)

    #
    #   SecureMessageDelegate
    #

    def decode_key(self, key: str, msg: SecureMessage) -> Optional[bytes]:
        assert not is_broadcast(msg=msg), 'broadcast message has no key'
        return base64_decode(string=key)

    # noinspection PyUnusedLocal
    def decrypt_key(self, data: bytes, sender: ID, receiver: ID, msg: SecureMessage) -> Optional[bytes]:
        assert not is_broadcast(msg=msg), 'broadcast message has no key'
        # decrypt key data with the receiver's private key
        user = self.barrack.user(identifier=msg.receiver)
        assert user is not None, 'failed to create local user: %s' % msg.receiver
        return user.decrypt(data=data)

    def deserialize_key(self, data: Optional[bytes], sender: ID, receiver: ID,
                        msg: SecureMessage) -> Optional[SymmetricKey]:
        if data is None:
            # get key from cache
            return self.key_cache.cipher_key(sender=sender, receiver=receiver)
        else:
            assert not is_broadcast(msg=msg), 'broadcast message has no key: %s' % msg
            dictionary = json_decode(data=data)
            # TODO: translate short keys
            #       'A' -> 'algorithm'
            #       'D' -> 'data'
            #       'V' -> 'iv'
            #       'M' -> 'mode'
            #       'P' -> 'padding'
            return SymmetricKey.parse(key=dictionary)

    def decode_data(self, data: str, msg: SecureMessage) -> Optional[bytes]:
        if is_broadcast(msg=msg):
            # broadcast message content will not be encrypted (just encoded to JsON),
            # so return the string data directly
            return utf8_encode(string=data)
        return base64_decode(string=data)

    def decrypt_content(self, data: bytes, key: SymmetricKey, msg: SecureMessage) -> Optional[bytes]:
        return key.decrypt(data)

    def deserialize_content(self, data: bytes, key: SymmetricKey, msg: SecureMessage) -> Optional[Content]:
        dictionary = json_decode(data=data)
        # TODO: translate short keys
        #       'T' -> 'type'
        #       'N' -> 'sn'
        #       'G' -> 'group'
        content = Content.parse(content=dictionary)
        assert content is not None, 'content error: %d' % len(data)

        if not is_broadcast(msg=msg):
            # check and cache key for reuse
            group = self.overt_group(content=content)
            if group is None:
                # personal message or (group) command
                # cache key with direction (sender -> receiver)
                self.key_cache.cache_cipher_key(key=key, sender=msg.sender, receiver=msg.receiver)
            else:
                # group message (excludes group command)
                # cache the key with direction (sender -> group)
                self.key_cache.cache_cipher_key(key=key, sender=msg.sender, receiver=group)
        # NOTICE: check attachment for File/Image/Audio/Video message content
        #         after deserialize content, this job should be do in subclass
        return content

    # noinspection PyUnusedLocal
    def sign_data(self, data: bytes, sender: ID, msg: SecureMessage) -> bytes:
        user = self.barrack.user(identifier=sender)
        assert user is not None, 'failed to sign with sender: %s' % sender
        return user.sign(data)

    def encode_signature(self, signature: bytes, msg: SecureMessage) -> str:
        return base64_encode(data=signature)

    #
    #   ReliableMessageDelegate
    #

    def decode_signature(self, signature: str, msg: ReliableMessage) -> Optional[bytes]:
        return base64_decode(string=signature)

    def verify_data_signature(self, data: bytes, signature: bytes, sender: ID, msg: ReliableMessage) -> bool:
        contact = self.barrack.user(identifier=sender)
        assert contact is not None, 'failed to verify signature for sender: %s' % sender
        return contact.verify(data=data, signature=signature)


def is_broadcast(msg: Message) -> bool:
    receiver = msg.group
    if receiver is None:
        receiver = msg.receiver
    return receiver.is_broadcast



# -*- coding: utf-8 -*-
#
#   DIMP : Decentralized Instant Messaging Protocol
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Group for assembly
    ~~~~~~~~~~~~~~~~~~

    Group with members
"""

from abc import abstractmethod
from typing import Optional, List

from mkm import ID, Bulletin, Document

from .entity import Entity, EntityDataSource


class GroupDataSource(EntityDataSource):
    """This interface is for getting information for group

        Group Data Source
        ~~~~~~~~~~~~~~~~~

        1. founder has the same public key with the group's meta.key
        2. owner and members should be set complying with the consensus algorithm
    """

    @abstractmethod
    def founder(self, identifier: ID) -> Optional[ID]:
        """
        Get founder of the group

        :param identifier: group ID
        :return: founder ID
        """
        raise NotImplemented

    @abstractmethod
    def owner(self, identifier: ID) -> Optional[ID]:
        """
        Get current owner of the group

        :param identifier: group ID
        :return: owner ID
        """
        raise NotImplemented

    @abstractmethod
    def members(self, identifier: ID) -> Optional[List[ID]]:
        """
        Get all members in the group

        :param identifier: group ID
        :return: member ID list
        """
        raise NotImplemented

    @abstractmethod
    def assistants(self, identifier: ID) -> Optional[List[ID]]:
        """
        Get assistants for this group

        :param identifier: group ID
        :return: robot ID list
        """
        raise NotImplemented


class Group(Entity):
    """This class is for creating group

        Group for organizing users
        ~~~~~~~~~~~~~~~~~~~~~~~~~~

            roles:
                founder
                owner
                members
                administrators - Optional
                assistants     - Optional
    """

    def __init__(self, identifier: ID):
        super().__init__(identifier=identifier)
        # once the group founder is set, it will never change
        self.__founder: ID = None

    @Entity.delegate.getter
    def delegate(self) -> Optional[GroupDataSource]:
        return super().delegate

    # @delegate.setter
    # def delegate(self, value: GroupDataSource):
    #     super(Group, Group).delegate.__set__(self, value)

    @property
    def bulletin(self) -> Optional[Bulletin]:
        doc = self.document(doc_type=Document.BULLETIN)
        if isinstance(doc, Bulletin):
            return doc

    @property
    def founder(self) -> ID:
        if self.__founder is None:
            assert isinstance(self.delegate, GroupDataSource), 'group data source error: %s' % self.delegate
            self.__founder = self.delegate.founder(identifier=self.identifier)
        return self.__founder

    @property
    def owner(self) -> ID:
        assert isinstance(self.delegate, GroupDataSource), 'group data source error: %s' % self.delegate
        return self.delegate.owner(identifier=self.identifier)

    @property
    def members(self) -> List[ID]:
        """
        NOTICE: the owner must be a member
                (usually the first one)

        :return: members ID list
        """
        assert isinstance(self.delegate, GroupDataSource), 'group data source error: %s' % self.delegate
        return self.delegate.members(identifier=self.identifier)

    @property
    def assistants(self) -> List[ID]:
        assert isinstance(self.delegate, GroupDataSource), 'group data source error: %s' % self.delegate
        return self.delegate.assistants(identifier=self.identifier)
